#   Copyright 2021 Huawei, Inc. All rights reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Node Local IP action implementations"""

import argparse
import logging
from collections.abc import Iterable, Sequence
from typing import Any

from openstack.network.v2 import local_ip_association as _local_ip_association
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common

LOG = logging.getLogger(__name__)


def _get_columns(
    item: _local_ip_association.LocalIPAssociation,
) -> tuple[tuple[str, ...], tuple[str, ...]]:
    hidden_columns = ['location', 'name', 'id', 'tenant_id']
    return utils.get_osc_show_columns_for_sdk_resource(
        item, {}, hidden_columns
    )


class CreateLocalIPAssociation(command.ShowOne):
    _description = _("Create Local IP Association")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'local_ip',
            metavar='<local-ip>',
            help=_(
                "Local IP that the port association belongs to (Name or ID)"
            ),
        )
        parser.add_argument(
            'fixed_port',
            metavar='<fixed-port>',
            help=_("The ID or Name of Port to allocate Local IP Association"),
        )
        parser.add_argument(
            '--fixed-ip',
            metavar='<fixed-ip>',
            help=_("Fixed IP for Local IP Association"),
        )

        identity_common.add_project_domain_option_to_parser(parser)

        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network

        attrs = {}
        port = client.find_port(parsed_args.fixed_port, ignore_missing=False)
        attrs['fixed_port_id'] = port.id
        if parsed_args.fixed_ip:
            attrs['fixed_ip'] = parsed_args.fixed_ip
        local_ip = client.find_local_ip(
            parsed_args.local_ip,
            ignore_missing=False,
        )
        obj = client.create_local_ip_association(local_ip.id, **attrs)
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns, formatters={})

        return (display_columns, data)


class DeleteLocalIPAssociation(command.Command):
    _description = _("Delete Local IP association(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'local_ip',
            metavar="<local-ip>",
            help=_(
                "Local IP that the port association belongs to (Name or ID)"
            ),
        )
        parser.add_argument(
            'fixed_port_id',
            nargs="+",
            metavar="<fixed-port-id>",
            help=_("The fixed port ID of Local IP Association"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        local_ip = client.find_local_ip(
            parsed_args.local_ip,
            ignore_missing=False,
        )
        result = 0

        for fixed_port_id in parsed_args.fixed_port_id:
            try:
                client.delete_local_ip_association(
                    local_ip.id,
                    fixed_port_id,
                    ignore_missing=False,
                )
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete Local IP Association with "
                        "fixed port "
                        "name or ID '%(fixed_port_id)s': %(e)s"
                    ),
                    {'fixed_port_id': fixed_port_id, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.fixed_port_id)
            msg = _(
                "%(result)s of %(total)s Local IP Associations failed "
                "to delete."
            ) % {'result': result, 'total': total}
            raise exceptions.CommandError(msg)


class ListLocalIPAssociation(command.Lister):
    _description = _("List Local IP Associations")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'local_ip',
            metavar='<local-ip>',
            help=_("Local IP that port associations belongs to"),
        )
        parser.add_argument(
            '--fixed-port',
            metavar='<fixed-port>',
            help=_(
                "List only local IP assocations with the specified fixed IP "
                "port (name or ID)"
            ),
        )
        parser.add_argument(
            '--fixed-ip',
            metavar='<fixed-ip>',
            help=_(
                "List only local IP associations with the specified fixed IP"
            ),
        )
        parser.add_argument(
            '--host',
            metavar='<host>',
            help=_("List only local IP associations with the specified host"),
        )
        identity_common.add_project_domain_option_to_parser(parser)

        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        client = self.app.client_manager.network
        columns = (
            'local_ip_id',
            'local_ip_address',
            'fixed_port_id',
            'fixed_ip',
            'host',
        )
        column_headers = (
            'Local IP ID',
            'Local IP Address',
            'Fixed port ID',
            'Fixed IP',
            'Host',
        )
        attrs = {}
        obj = client.find_local_ip(
            parsed_args.local_ip,
            ignore_missing=False,
        )
        if parsed_args.fixed_port:
            port = client.find_port(
                parsed_args.fixed_port, ignore_missing=False
            )
            attrs['fixed_port_id'] = port.id
        if parsed_args.fixed_ip:
            attrs['fixed_ip'] = parsed_args.fixed_ip
        if parsed_args.host:
            attrs['host'] = parsed_args.host

        data = client.local_ip_associations(obj, **attrs)

        return (
            column_headers,
            (
                utils.get_item_properties(s, columns, formatters={})
                for s in data
            ),
        )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""Compute v2 API Library

A collection of wrappers for deprecated Compute v2 APIs that are not
intentionally supported by SDK. Most of these are proxy APIs.
"""

import http
from typing import Any

from openstack import exceptions as sdk_exceptions
from osc_lib import exceptions


# security groups


def create_security_group(
    compute_client: Any,
    name: str | None = None,
    description: str | None = None,
) -> Any:
    """Create a new security group

    https://docs.openstack.org/api-ref/compute/#create-security-group

    :param compute_client: A compute client
    :param str name: Security group name
    :param str description: Security group description
    :returns: A security group object
    """
    data = {
        'name': name,
        'description': description,
    }
    response = compute_client.post(
        '/os-security-groups', data=data, microversion='2.1'
    )
    sdk_exceptions.raise_from_response(response)
    return response.json()['security_group']


def list_security_groups(
    compute_client: Any, all_projects: bool | None = None
) -> Any:
    """Get all security groups

    https://docs.openstack.org/api-ref/compute/#list-security-groups

    :param compute_client: A compute client
    :param bool all_projects: If true, list from all projects
    :returns: A list of security group objects
    """
    url = '/os-security-groups'
    if all_projects is not None:
        url += f'?all_tenants={all_projects}'
    response = compute_client.get(url, microversion='2.1')
    sdk_exceptions.raise_from_response(response)
    return response.json()['security_groups']


def find_security_group(compute_client: Any, name_or_id: str) -> Any:
    """Find the security group for a given name or ID

    https://docs.openstack.org/api-ref/compute/#show-security-group-details

    :param compute_client: A compute client
    :param name_or_id: The name or ID of the security group to look up
    :returns: A security group object
    :raises exception.NotFound: If a matching security group could not be
        found or more than one match was found
    """
    response = compute_client.get(
        f'/os-security-groups/{name_or_id}', microversion='2.1'
    )
    if response.status_code != http.HTTPStatus.NOT_FOUND:
        # there might be other, non-404 errors
        sdk_exceptions.raise_from_response(response)
        return response.json()['security_group']

    response = compute_client.get('/os-security-groups', microversion='2.1')
    sdk_exceptions.raise_from_response(response)
    found = None
    security_groups = response.json()['security_groups']
    for security_group in security_groups:
        if security_group['name'] == name_or_id:
            if found:
                raise exceptions.NotFound(
                    f'multiple matches found for {name_or_id}'
                )
            found = security_group

    if not found:
        raise exceptions.NotFound(f'{name_or_id} not found')

    return found


def update_security_group(
    compute_client: Any,
    security_group_id: str,
    name: str | None = None,
    description: str | None = None,
) -> Any:
    """Update an existing security group

    https://docs.openstack.org/api-ref/compute/#update-security-group

    :param compute_client: A compute client
    :param str security_group_id: The ID of the security group to update
    :param str name: Security group name
    :param str description: Security group description
    :returns: A security group object
    """
    data = {}
    if name:
        data['name'] = name
    if description:
        data['description'] = description
    response = compute_client.put(
        f'/os-security-groups/{security_group_id}',
        data=data,
        microversion='2.1',
    )
    sdk_exceptions.raise_from_response(response)
    return response.json()['security_group']


def delete_security_group(
    compute_client: Any, security_group_id: str | None = None
) -> None:
    """Delete a security group

    https://docs.openstack.org/api-ref/compute/#delete-security-group

    :param compute_client: A compute client
    :param str security_group_id: Security group ID
    :returns: None
    """
    response = compute_client.delete(
        f'/os-security-groups/{security_group_id}', microversion='2.1'
    )
    sdk_exceptions.raise_from_response(response)


# security group rules


def create_security_group_rule(
    compute_client: Any,
    security_group_id: str | None = None,
    ip_protocol: str | None = None,
    from_port: int | None = None,
    to_port: int | None = None,
    remote_ip: str | None = None,
    remote_group: str | None = None,
) -> Any:
    """Create a new security group rule

    https://docs.openstack.org/api-ref/compute/#create-security-group-rule

    :param compute_client: A compute client
    :param str security_group_id: Security group ID
    :param str ip_protocol: IP protocol, 'tcp', 'udp' or 'icmp'
    :param int from_port: Source port
    :param int to_port: Destination port
    :param str remote_ip: Source IP address in CIDR notation
    :param str remote_group: Remote security group
    :returns: A security group object
    """
    data = {
        'parent_group_id': security_group_id,
        'ip_protocol': ip_protocol,
        'from_port': from_port,
        'to_port': to_port,
        'cidr': remote_ip,
        'group_id': remote_group,
    }
    response = compute_client.post(
        '/os-security-group-rules', data=data, microversion='2.1'
    )
    sdk_exceptions.raise_from_response(response)
    return response.json()['security_group_rule']


def delete_security_group_rule(
    compute_client: Any, security_group_rule_id: str | None = None
) -> None:
    """Delete a security group rule

    https://docs.openstack.org/api-ref/compute/#delete-security-group-rule

    :param compute_client: A compute client
    :param str security_group_rule_id: Security group rule ID
    :returns: None
    """
    response = compute_client.delete(
        f'/os-security-group-rules/{security_group_rule_id}',
        microversion='2.1',
    )
    sdk_exceptions.raise_from_response(response)


# networks


def create_network(
    compute_client: Any,
    name: str,
    subnet: str,
    share_subnet: bool | None = None,
) -> Any:
    """Create a new network

    https://docs.openstack.org/api-ref/compute/#create-network

    :param compute_client: A compute client
    :param str name: Network label
    :param int subnet: Subnet for IPv4 fixed addresses in CIDR notation
    :param bool share_subnet: Shared subnet between projects
    :returns: A network object
    """
    data: dict[str, Any] = {
        'label': name,
        'cidr': subnet,
    }
    if share_subnet is not None:
        data['share_address'] = share_subnet

    response = compute_client.post(
        '/os-networks', data=data, microversion='2.1'
    )
    sdk_exceptions.raise_from_response(response)
    return response.json()['network']


def list_networks(compute_client: Any) -> Any:
    """Get all networks

    https://docs.openstack.org/api-ref/compute/#list-networks

    :param compute_client: A compute client
    :returns: A list of network objects
    """
    response = compute_client.get('/os-networks', microversion='2.1')
    sdk_exceptions.raise_from_response(response)
    return response.json()['networks']


def find_network(compute_client: Any, name_or_id: str) -> Any:
    """Find the network for a given name or ID

    https://docs.openstack.org/api-ref/compute/#show-network-details

    :param compute_client: A compute client
    :param name_or_id: The name or ID of the network to look up
    :returns: A network object
    :raises exception.NotFound: If a matching network could not be found or
        more than one match was found
    """
    response = compute_client.get(
        f'/os-networks/{name_or_id}', microversion='2.1'
    )
    if response.status_code != http.HTTPStatus.NOT_FOUND:
        # there might be other, non-404 errors
        sdk_exceptions.raise_from_response(response)
        return response.json()['network']

    response = compute_client.get('/os-networks', microversion='2.1')
    sdk_exceptions.raise_from_response(response)
    found = None
    networks = response.json()['networks']
    for network in networks:
        if network['label'] == name_or_id:
            if found:
                raise exceptions.NotFound(
                    f'multiple matches found for {name_or_id}'
                )
            found = network

    if not found:
        raise exceptions.NotFound(f'{name_or_id} not found')

    return found


def delete_network(compute_client: Any, network_id: str) -> None:
    """Delete a network

    https://docs.openstack.org/api-ref/compute/#delete-network

    :param compute_client: A compute client
    :param string network_id: The network ID
    :returns: None
    """
    response = compute_client.delete(
        f'/os-networks/{network_id}', microversion='2.1'
    )
    sdk_exceptions.raise_from_response(response)


# floating ips


def create_floating_ip(compute_client: Any, network: str) -> Any:
    """Create a new floating ip

    https://docs.openstack.org/api-ref/compute/#create-allocate-floating-ip-address

    :param network: Name of floating IP pool
    :returns: A floating IP object
    """
    response = compute_client.post(
        '/os-floating-ips', data={'pool': network}, microversion='2.1'
    )
    sdk_exceptions.raise_from_response(response)
    return response.json()['floating_ip']


def list_floating_ips(compute_client: Any) -> Any:
    """Get all floating IPs

    https://docs.openstack.org/api-ref/compute/#list-floating-ip-addresses

    :returns: A list of floating IP objects
    """
    response = compute_client.get('/os-floating-ips', microversion='2.1')
    sdk_exceptions.raise_from_response(response)
    return response.json()['floating_ips']


def get_floating_ip(compute_client: Any, floating_ip_id: str) -> Any:
    """Get a floating IP

    https://docs.openstack.org/api-ref/compute/#show-floating-ip-address-details

    :param string floating_ip_id: The floating IP address
    :returns: A floating IP object
    """
    response = compute_client.get(
        f'/os-floating-ips/{floating_ip_id}', microversion='2.1'
    )
    sdk_exceptions.raise_from_response(response)
    return response.json()['floating_ip']


def delete_floating_ip(compute_client: Any, floating_ip_id: str) -> None:
    """Delete a floating IP

    https://docs.openstack.org/api-ref/compute/#delete-deallocate-floating-ip-address

    :param string floating_ip_id: The floating IP address
    :returns: None
    """
    response = compute_client.delete(
        f'/os-floating-ips/{floating_ip_id}', microversion='2.1'
    )
    sdk_exceptions.raise_from_response(response)


# floating ip pools


def list_floating_ip_pools(compute_client: Any) -> Any:
    """Get all floating IP pools

    https://docs.openstack.org/api-ref/compute/#list-floating-ip-pools

    :param compute_client: A compute client
    :returns: A list of floating IP pool objects
    """
    response = compute_client.get('/os-floating-ip-pools', microversion='2.1')
    sdk_exceptions.raise_from_response(response)

    return response.json()['floating_ip_pools']



# Copyright 2017 FUJITSU LIMITED
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from osc_lib.cli import format_columns
from osc_lib import exceptions

from openstack.identity.v3 import project as _project
from openstack.network.v2 import vpn_endpoint_group as _vpn_endpoint_group
from openstack.network.v2 import vpn_ike_policy as _vpn_ike_policy
from openstack.network.v2 import vpn_ipsec_policy as sdk_vpn_ipsecp_p
from openstack.network.v2 import vpn_ipsec_site_connection as sdk_vpn_ipsec_sc
from openstack.network.v2 import vpn_service as sdk_vpn_service
from openstack.test import fakes as sdk_fakes

from openstackclient.network.v2.vpnaas import ipsec_site_connection
from openstackclient.tests.unit.network.v2 import fakes as test_fakes
from openstackclient.tests.unit.network.v2.vpnaas import fakes
from openstackclient.tests.unit import utils as tests_utils


class TestIPsecSiteConn(test_fakes.TestNetworkV2):
    def setUp(self):
        super().setUp()

        self.project = sdk_fakes.generate_fake_resource(_project.Project)
        self.identity_sdk_client.find_project.return_value = self.project

        self._ipsec_site_conn = fakes.IPsecSiteConnection().create_conn()

        def _mock_ipsec_site_conn(*args, **kwargs):
            return {'id': args[0]}

        self.network_client.find_vpn_ipsec_site_connection.side_effect = (
            mock.Mock(side_effect=_mock_ipsec_site_conn)
        )
        self.headers = (
            'ID',
            'Name',
            'Peer Address',
            'Authentication Algorithm',
            'Status',
            'Project',
            'Peer CIDRs',
            'VPN Service',
            'IPSec Policy',
            'IKE Policy',
            'MTU',
            'Initiator',
            'State',
            'Description',
            'Pre-shared Key',
            'Route Mode',
            'Local ID',
            'Peer ID',
            'Local Endpoint Group ID',
            'Peer Endpoint Group ID',
            'DPD',
        )
        self.ordered_headers = (
            'Authentication Algorithm',
            'DPD',
            'Description',
            'ID',
            'IKE Policy',
            'IPSec Policy',
            'Initiator',
            'Local Endpoint Group ID',
            'Local ID',
            'MTU',
            'Name',
            'Peer Address',
            'Peer CIDRs',
            'Peer Endpoint Group ID',
            'Peer ID',
            'Pre-shared Key',
            'Project',
            'Route Mode',
            'State',
            'Status',
            'VPN Service',
        )
        self.ordered_data = (
            self._ipsec_site_conn['auth_mode'],
            self._ipsec_site_conn['dpd'],
            self._ipsec_site_conn['description'],
            self._ipsec_site_conn['id'],
            self._ipsec_site_conn['ikepolicy_id'],
            self._ipsec_site_conn['ipsecpolicy_id'],
            self._ipsec_site_conn['initiator'],
            self._ipsec_site_conn['local_ep_group_id'],
            self._ipsec_site_conn['local_id'],
            self._ipsec_site_conn['mtu'],
            self._ipsec_site_conn['name'],
            self._ipsec_site_conn['peer_address'],
            format_columns.ListColumn(self._ipsec_site_conn['peer_cidrs']),
            self._ipsec_site_conn['peer_ep_group_id'],
            self._ipsec_site_conn['peer_id'],
            self._ipsec_site_conn['psk'],
            self._ipsec_site_conn['project_id'],
            self._ipsec_site_conn['route_mode'],
            self._ipsec_site_conn['admin_state_up'],
            self._ipsec_site_conn['status'],
            self._ipsec_site_conn['vpnservice_id'],
        )


class TestCreateIPsecSiteConn(TestIPsecSiteConn):
    def setUp(self):
        super().setUp()

        self.project = sdk_fakes.generate_fake_resource(_project.Project)
        self.identity_sdk_client.find_project.return_value = self.project

        self.network_client.create_vpn_ipsec_site_connection.return_value = (
            self._ipsec_site_conn
        )
        self.mocked = self.network_client.create_vpn_ipsec_site_connection
        self.cmd = ipsec_site_connection.CreateIPsecSiteConnection(
            self.app, None
        )

    def test_create_with_no_options(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_create_with_all_params(self):
        arglist = [
            '--project',
            self.project.id,
            '--peer-address',
            '192.168.2.10',
            '--peer-id',
            '192.168.2.10',
            '--psk',
            'abcd',
            '--initiator',
            'bi-directional',
            '--vpnservice',
            'vpnservice_id',
            '--ikepolicy',
            'ikepolicy_id',
            '--ipsecpolicy',
            'ipsecpolicy_id',
            '--mtu',
            '1500',
            '--description',
            'my-vpn-connection',
            '--local-endpoint-group',
            'local-epg',
            '--peer-endpoint-group',
            'peer-epg',
            self._ipsec_site_conn.name,
        ]
        verifylist = [
            ('project', self.project.id),
            ('peer_address', '192.168.2.10'),
            ('peer_id', '192.168.2.10'),
            ('psk', 'abcd'),
            ('initiator', 'bi-directional'),
            ('vpnservice', 'vpnservice_id'),
            ('ikepolicy', 'ikepolicy_id'),
            ('ipsecpolicy', 'ipsecpolicy_id'),
            ('mtu', '1500'),
            ('description', 'my-vpn-connection'),
            ('local_endpoint_group', 'local-epg'),
            ('peer_endpoint_group', 'peer-epg'),
            ('name', self._ipsec_site_conn.name),
        ]

        def _mock_endpoint_group(*args, **kwargs):
            return _vpn_endpoint_group.VpnEndpointGroup(**{'id': args[0]})

        def _mock_vpn_service(*args, **kwargs):
            return sdk_vpn_service.VpnService(**{'id': args[0]})

        def _mock_vpn_ike_policy(*args, **kwargs):
            return _vpn_ike_policy.VpnIkePolicy(**{'id': args[0]})

        def _mock_vpn_ipsec_policy(*args, **kwargs):
            return sdk_vpn_ipsecp_p.VpnIpsecPolicy(**{'id': args[0]})

        self.network_client.find_vpn_endpoint_group.side_effect = mock.Mock(
            side_effect=_mock_endpoint_group
        )
        self.network_client.find_vpn_service.side_effect = mock.Mock(
            side_effect=_mock_vpn_service
        )
        self.network_client.find_vpn_ike_policy.side_effect = mock.Mock(
            side_effect=_mock_vpn_ike_policy
        )
        self.network_client.find_vpn_ipsec_policy.side_effect = mock.Mock(
            side_effect=_mock_vpn_ipsec_policy
        )
        self.network_client.create_vpn_ipsec_site_connection.return_value = (
            self._ipsec_site_conn
        )
        self.ordered_data = (
            self._ipsec_site_conn['auth_mode'],
            self._ipsec_site_conn['dpd'],
            self._ipsec_site_conn['description'],
            self._ipsec_site_conn['id'],
            self._ipsec_site_conn['ikepolicy_id'],
            self._ipsec_site_conn['ipsecpolicy_id'],
            self._ipsec_site_conn['initiator'],
            self._ipsec_site_conn['local_ep_group_id'],
            self._ipsec_site_conn['local_id'],
            self._ipsec_site_conn['mtu'],
            self._ipsec_site_conn['name'],
            self._ipsec_site_conn['peer_address'],
            format_columns.ListColumn(self._ipsec_site_conn['peer_cidrs']),
            self._ipsec_site_conn['peer_ep_group_id'],
            self._ipsec_site_conn['peer_id'],
            self._ipsec_site_conn['psk'],
            self._ipsec_site_conn['project_id'],
            self._ipsec_site_conn['route_mode'],
            self._ipsec_site_conn['admin_state_up'],
            self._ipsec_site_conn['status'],
            self._ipsec_site_conn['vpnservice_id'],
        )
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with(
            project_id=self.project.id,
            peer_address='192.168.2.10',
            peer_id='192.168.2.10',
            psk='abcd',
            initiator='bi-directional',
            vpnservice_id='vpnservice_id',
            ikepolicy_id='ikepolicy_id',
            ipsecpolicy_id='ipsecpolicy_id',
            mtu='1500',
            description='my-vpn-connection',
            local_ep_group_id='local-epg',
            peer_ep_group_id='peer-epg',
            name=self._ipsec_site_conn.name,
        )
        self.assertEqual(self.ordered_headers, tuple(sorted(headers)))
        self.assertCountEqual(self.ordered_data, data)


class TestDeleteIPsecSiteConn(TestIPsecSiteConn):
    def setUp(self):
        super().setUp()
        self.cmd = ipsec_site_connection.DeleteIPsecSiteConnection(
            self.app, None
        )

    def test_delete_with_one_resource(self):
        def _mock_ips_sc(*args, **kwargs):
            return sdk_vpn_ipsec_sc.VpnIPSecSiteConnection(**{'id': args[0]})

        self.network_client.find_vpn_endpoint_group.side_effect = _mock_ips_sc

        arglist = [self._ipsec_site_conn['id']]
        verifylist = [('ipsec_site_connection', [self._ipsec_site_conn['id']])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.delete_vpn_ipsec_site_connection.assert_called_once_with(
            self._ipsec_site_conn['id']
        )
        self.assertIsNone(result)

    def test_delete_with_multiple_resources(self):

        def _mock_ips_c(*args, **kwargs):
            return sdk_vpn_ipsec_sc.VpnIPSecSiteConnection(**{'id': args[0]})

        self.network_client.find_vpn_ipsec_policy.side_effect = _mock_ips_c

        arglist = ['target1', 'target2']
        verifylist = [('ipsec_site_connection', ['target1', 'target2'])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.assertEqual(
            2,
            self.network_client.delete_vpn_ipsec_site_connection.call_count,
        )
        for idx, reference in enumerate(['target1', 'target2']):
            actual = ''.join(
                self.network_client.delete_vpn_ipsec_site_connection.call_args_list[
                    idx
                ][0]
            )
            self.assertEqual(reference, actual)

    def test_delete_multiple_with_exception(self):
        arglist = ['target']
        verifylist = [('ipsec_site_connection', ['target'])]

        self.network_client.find_vpn_ipsec_site_connection.side_effect = [
            'target',
            exceptions.CommandError,
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestListIPsecSiteConn(TestIPsecSiteConn):
    def setUp(self):
        super().setUp()
        self.cmd = ipsec_site_connection.ListIPsecSiteConnection(
            self.app, None
        )

        self.short_header = (
            'ID',
            'Name',
            'Peer Address',
            'Authentication Algorithm',
            'Status',
        )

        self.short_data = (
            self._ipsec_site_conn['id'],
            self._ipsec_site_conn['name'],
            self._ipsec_site_conn['peer_address'],
            self._ipsec_site_conn['auth_mode'],
            self._ipsec_site_conn['status'],
        )

        self.network_client.vpn_ipsec_site_connections.return_value = [
            self._ipsec_site_conn
        ]

    def test_list_with_long_option(self):
        arglist = ['--long']
        verifylist = [('long', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, _data = self.cmd.take_action(parsed_args)

        self.network_client.vpn_ipsec_site_connections.assert_called_once_with()
        self.assertEqual(list(self.headers), headers)

    def test_list_with_no_option(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)

        self.network_client.vpn_ipsec_site_connections.assert_called_once_with()
        self.assertEqual(list(self.short_header), headers)
        self.assertEqual([self.short_data], list(data))


class TestSetIPsecSiteConn(TestIPsecSiteConn):
    def setUp(self):
        super().setUp()

        self.network_client.update_vpn_ipsec_site_connection.return_value = (
            self._ipsec_site_conn
        )

        self.cmd = ipsec_site_connection.SetIPsecSiteConnection(self.app, None)

    def test_set_ipsec_site_conn_with_peer_id(self):
        arglist = [
            self._ipsec_site_conn['id'],
            '--peer-id',
            '192.168.3.10',
        ]
        verifylist = [
            ('ipsec_site_connection', self._ipsec_site_conn['id']),
            ('peer_id', '192.168.3.10'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.update_vpn_ipsec_site_connection.assert_called_once_with(
            self._ipsec_site_conn['id'], peer_id='192.168.3.10'
        )
        self.assertIsNone(result)

    def test_set_name(self):
        arglist = [
            self._ipsec_site_conn['id'],
            '--name',
            'change',
        ]
        verifylist = [
            ('ipsec_site_connection', self._ipsec_site_conn['id']),
            ('name', 'change'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.update_vpn_ipsec_site_connection.assert_called_once_with(
            self._ipsec_site_conn['id'], name='change'
        )
        self.assertIsNone(result)

    def test_set_description(self):
        arglist = [
            self._ipsec_site_conn['id'],
            '--description',
            'change-desc',
        ]
        verifylist = [
            ('ipsec_site_connection', self._ipsec_site_conn['id']),
            ('description', 'change-desc'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.update_vpn_ipsec_site_connection.assert_called_once_with(
            self._ipsec_site_conn['id'], description='change-desc'
        )
        self.assertIsNone(result)


class TestShowIPsecSiteConn(TestIPsecSiteConn):
    def setUp(self):
        super().setUp()

        self.network_client.find_vpn_ipsec_site_connection.side_effect = None
        self.network_client.find_vpn_ipsec_site_connection.return_value = (
            self._ipsec_site_conn
        )

        self.cmd = ipsec_site_connection.ShowIPsecSiteConnection(
            self.app, None
        )

    def test_show_filtered_by_id_or_name(self):
        arglist = [
            self._ipsec_site_conn['id'],
        ]
        verifylist = [
            ('ipsec_site_connection', self._ipsec_site_conn['id']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)

        self.network_client.find_vpn_ipsec_site_connection.assert_called_once_with(
            self._ipsec_site_conn['id'], ignore_missing=False
        )
        self.assertEqual(self.ordered_headers, headers)
        self.assertCountEqual(self.ordered_data, data)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import datetime
from unittest import mock

from openstack.compute.v2 import usage as _usage
from openstack.identity.v3 import project as _project
from openstack.test import fakes as sdk_fakes

from openstackclient.compute.v2 import usage as usage_cmds
from openstackclient.tests.unit.compute.v2 import fakes as compute_fakes


class TestUsageList(compute_fakes.TestCompute):
    def setUp(self):
        super().setUp()

        self.project = sdk_fakes.generate_fake_resource(_project.Project)
        self.identity_sdk_client.projects.return_value = [self.project]

        self.usages = [
            sdk_fakes.generate_fake_resource(
                _usage.Usage, project_id=self.project.name
            )
        ]
        self.columns = (
            "Project",
            "Servers",
            "RAM MB-Hours",
            "CPU Hours",
            "Disk GB-Hours",
        )
        self.data = [
            (
                usage_cmds.ProjectColumn(self.usages[0].project_id),
                usage_cmds.CountColumn(self.usages[0].server_usages),
                usage_cmds.FloatColumn(self.usages[0].total_memory_mb_usage),
                usage_cmds.FloatColumn(self.usages[0].total_vcpus_usage),
                usage_cmds.FloatColumn(self.usages[0].total_local_gb_usage),
            )
        ]

        self.compute_client.usages.return_value = self.usages

        # Get the command object to test
        self.cmd = usage_cmds.ListUsage(self.app, None)

    def test_usage_list_no_options(self):
        arglist = []
        verifylist = [
            ('start', None),
            ('end', None),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.projects.assert_called_once()

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(tuple(self.data), tuple(data))

    def test_usage_list_with_options(self):
        arglist = [
            '--start',
            '2016-11-11',
            '--end',
            '2016-12-20',
        ]
        verifylist = [
            ('start', '2016-11-11'),
            ('end', '2016-12-20'),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.projects.assert_called_once()
        self.compute_client.usages.assert_called_with(
            start=datetime.datetime(2016, 11, 11, 0, 0),
            end=datetime.datetime(2016, 12, 20, 0, 0),
            detailed=True,
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(tuple(self.data), tuple(data))

    def test_usage_list_with_pagination(self):
        arglist = []
        verifylist = [
            ('start', None),
            ('end', None),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.projects.assert_called_once()
        self.compute_client.usages.assert_has_calls(
            [mock.call(start=mock.ANY, end=mock.ANY, detailed=True)]
        )
        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(tuple(self.data), tuple(data))


class TestUsageShow(compute_fakes.TestCompute):
    # Return value of self.usage_mock.list().
    def setUp(self):
        super().setUp()

        self.project = sdk_fakes.generate_fake_resource(_project.Project)
        self.identity_sdk_client.find_project.return_value = self.project

        self.usage = sdk_fakes.generate_fake_resource(
            _usage.Usage, project_id=self.project.name
        )
        self.compute_client.get_usage.return_value = self.usage

        self.columns = (
            'Project',
            'Servers',
            'RAM MB-Hours',
            'CPU Hours',
            'Disk GB-Hours',
        )
        self.data = (
            usage_cmds.ProjectColumn(self.usage.project_id),
            usage_cmds.CountColumn(self.usage.server_usages),
            usage_cmds.FloatColumn(self.usage.total_memory_mb_usage),
            usage_cmds.FloatColumn(self.usage.total_vcpus_usage),
            usage_cmds.FloatColumn(self.usage.total_local_gb_usage),
        )

        # Get the command object to test
        self.cmd = usage_cmds.ShowUsage(self.app, None)

    def test_usage_show_no_options(self):
        self.app.client_manager.auth_ref = mock.Mock()
        self.app.client_manager.auth_ref.project_id = self.project.id

        arglist = []
        verifylist = [
            ('project', None),
            ('start', None),
            ('end', None),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_usage_show_with_options(self):
        arglist = [
            '--project',
            self.project.id,
            '--start',
            '2016-11-11',
            '--end',
            '2016-12-20',
        ]
        verifylist = [
            ('project', self.project.id),
            ('start', '2016-11-11'),
            ('end', '2016-12-20'),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.compute_client.get_usage.assert_called_with(
            project=self.project.id,
            start=datetime.datetime(2016, 11, 11, 0, 0),
            end=datetime.datetime(2016, 12, 20, 0, 0),
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Flavor action implementations"""

import argparse
import logging
from collections.abc import Iterable, Sequence
from typing import Any

from openstack.network.v2 import flavor as _flavor
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.common import pagination
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common
from openstackclient.network import common

LOG = logging.getLogger(__name__)


def _get_columns(
    item: _flavor.Flavor,
) -> tuple[tuple[str, ...], tuple[str, ...]]:
    column_map = {
        'is_enabled': 'enabled',
    }

    hidden_columns = ['location', 'tenant_id']
    return utils.get_osc_show_columns_for_sdk_resource(
        item, column_map, hidden_columns
    )


def _get_attrs(
    client_manager: Any, parsed_args: argparse.Namespace
) -> dict[str, Any]:
    attrs: dict[str, Any] = {}
    attrs['name'] = parsed_args.name
    attrs['service_type'] = parsed_args.service_type
    if parsed_args.description is not None:
        attrs['description'] = parsed_args.description
    if parsed_args.enable:
        attrs['enabled'] = True
    if parsed_args.disable:
        attrs['enabled'] = False
    if 'project' in parsed_args and parsed_args.project is not None:
        identity_client = client_manager.sdk_connection.identity
        project_id = identity_common.find_project_id_sdk(
            identity_client,
            parsed_args.project,
            parsed_args.project_domain,
        )
        attrs['project_id'] = project_id

    return attrs


class AddNetworkFlavorToProfile(command.Command):
    _description = _("Add a service profile to a network flavor")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'flavor', metavar="<flavor>", help=_("Network flavor (name or ID)")
        )
        parser.add_argument(
            'service_profile',
            metavar="<service-profile>",
            help=_("Service profile (ID only)"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        obj_flavor = client.find_flavor(
            parsed_args.flavor, ignore_missing=False
        )
        obj_service_profile = client.find_service_profile(
            parsed_args.service_profile, ignore_missing=False
        )
        client.associate_flavor_with_service_profile(
            obj_flavor, obj_service_profile
        )


class CreateNetworkFlavor(command.ShowOne, common.NeutronCommandWithExtraArgs):
    _description = _("Create new network flavor")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'name', metavar="<name>", help=_("Name for the flavor")
        )
        parser.add_argument(
            '--service-type',
            metavar="<service-type>",
            required=True,
            help=_(
                'Service type to which the flavor applies. For example: VPN '
                '(See openstack network service provider list for loaded '
                'examples.)'
            ),
        )
        parser.add_argument(
            '--description', help=_('Description for the flavor')
        )
        parser.add_argument(
            '--project',
            metavar="<project>",
            help=_("Owner's project (name or ID)"),
        )
        identity_common.add_project_domain_option_to_parser(parser)

        enable_group = parser.add_mutually_exclusive_group()
        enable_group.add_argument(
            '--enable',
            action='store_true',
            help=_("Enable the flavor (default)"),
        )
        enable_group.add_argument(
            '--disable', action='store_true', help=_("Disable the flavor")
        )

        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        attrs = _get_attrs(self.app.client_manager, parsed_args)
        attrs.update(
            self._parse_extra_properties(parsed_args.extra_properties)
        )
        obj = client.create_flavor(**attrs)
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns, formatters={})

        return (display_columns, data)


class DeleteNetworkFlavor(command.Command):
    _description = _("Delete network flavors")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'flavor',
            metavar='<flavor>',
            nargs='+',
            help=_('Flavor(s) to delete (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        result = 0

        for flavor in parsed_args.flavor:
            try:
                obj = client.find_flavor(flavor, ignore_missing=False)
                client.delete_flavor(obj)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete flavor with "
                        "name or ID '%(flavor)s': %(e)s"
                    ),
                    {"flavor": flavor, "e": e},
                )
        if result > 0:
            total = len(parsed_args.flavor)
            msg = _("%(result)s of %(total)s flavors failed to delete.") % {
                "result": result,
                "total": total,
            }
            raise exceptions.CommandError(msg)


class ListNetworkFlavor(command.Lister):
    _description = _("List network flavors")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        pagination.add_marker_pagination_option_to_parser(parser)
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        client = self.app.client_manager.network

        columns = ('id', 'name', 'is_enabled', 'service_type', 'description')
        column_headers = (
            'ID',
            'Name',
            'Enabled',
            'Service Type',
            'Description',
        )

        filters = {}
        if parsed_args.marker is not None:
            filters['marker'] = parsed_args.marker
        if parsed_args.limit is not None:
            filters['limit'] = parsed_args.limit
        if parsed_args.max_items is not None:
            filters['max_items'] = parsed_args.max_items

        data = client.flavors(**filters)
        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                )
                for s in data
            ),
        )


class RemoveNetworkFlavorFromProfile(command.Command):
    _description = _("Remove service profile from network flavor")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'flavor', metavar="<flavor>", help=_("Network flavor (name or ID)")
        )
        parser.add_argument(
            'service_profile',
            metavar="<service-profile>",
            help=_("Service profile (ID only)"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        obj_flavor = client.find_flavor(
            parsed_args.flavor, ignore_missing=False
        )
        obj_service_profile = client.find_service_profile(
            parsed_args.service_profile, ignore_missing=False
        )
        client.disassociate_flavor_from_service_profile(
            obj_flavor, obj_service_profile
        )


class SetNetworkFlavor(common.NeutronCommandWithExtraArgs):
    _description = _("Set network flavor properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'flavor',
            metavar="<flavor>",
            help=_("Flavor to update (name or ID)"),
        )
        parser.add_argument(
            '--description', help=_('Set network flavor description')
        )
        enable_group = parser.add_mutually_exclusive_group()
        enable_group.add_argument(
            '--disable', action='store_true', help=_("Disable network flavor")
        )
        enable_group.add_argument(
            '--enable', action='store_true', help=_("Enable network flavor")
        )
        parser.add_argument(
            '--name', metavar="<name>", help=_('Set flavor name')
        )

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        obj = client.find_flavor(parsed_args.flavor, ignore_missing=False)
        attrs = {}
        if parsed_args.name is not None:
            attrs['name'] = parsed_args.name
        if parsed_args.description is not None:
            attrs['description'] = parsed_args.description
        if parsed_args.enable:
            attrs['enabled'] = True
        if parsed_args.disable:
            attrs['enabled'] = False
        attrs.update(
            self._parse_extra_properties(parsed_args.extra_properties)
        )
        client.update_flavor(obj, **attrs)


class ShowNetworkFlavor(command.ShowOne):
    _description = _("Display network flavor details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'flavor',
            metavar='<flavor>',
            help=_('Flavor to display (name or ID)'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        obj = client.find_flavor(parsed_args.flavor, ignore_missing=False)
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns)
        return display_columns, data



#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import random
from unittest import mock
import uuid

from openstack.image import v2 as image_v2
from openstack.image.v2 import cache
from openstack.image.v2 import image
from openstack.image.v2 import member
from openstack.image.v2 import metadef_namespace
from openstack.image.v2 import metadef_object
from openstack.image.v2 import metadef_property
from openstack.image.v2 import metadef_resource_type
from openstack.image.v2 import service_info as _service_info
from openstack.image.v2 import task

from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes
from openstackclient.tests.unit import utils


class FakeClientMixin:
    def setUp(self):
        super().setUp()

        # TODO(stephenfin): Switch to spec_set once keystoneauth exposes
        # instance attributes as class attributes
        # https://review.opendev.org/c/openstack/keystoneauth/+/994090
        self.app.client_manager.image = mock.Mock(spec=image_v2.Proxy)
        self.image_client = self.app.client_manager.image


class TestImagev2(
    identity_fakes.FakeClientMixin,
    FakeClientMixin,
    utils.TestCommand,
): ...


def create_one_image(attrs=None):
    """Create a fake image.

    :param attrs: A dictionary with all attributes of image
    :type attrs: dict
    :return: A fake Image object.
    :rtype: `openstack.image.v2.image.Image`
    """
    attrs = attrs or {}

    # Set default attribute
    image_info = {
        'id': str(uuid.uuid4()),
        'name': 'image-name' + uuid.uuid4().hex,
        'owner_id': 'image-owner' + uuid.uuid4().hex,
        'is_protected': bool(random.choice([0, 1])),
        'visibility': random.choice(['public', 'private']),
        'tags': [uuid.uuid4().hex for r in range(2)],
    }

    # Overwrite default attributes if there are some attributes set
    image_info.update(attrs)

    return image.Image(**image_info)


def create_images(attrs=None, count=2):
    """Create multiple fake images.

    :param attrs: A dictionary with all attributes of image
    :type attrs: dict
    :param count: The number of images to be faked
    :type count: int
    :return: A list of fake Image objects
    :rtype: list
    """
    images = []
    for n in range(0, count):
        images.append(create_one_image(attrs))

    return images


def create_one_image_member(attrs=None):
    """Create a fake image member.

    :param attrs: A dictionary with all attributes of image member
    :type attrs: dict
    :return: A fake Member object.
    :rtype: `openstack.image.v2.member.Member`
    """
    attrs = attrs or {}

    # Set default attribute
    image_member_info = {
        'member_id': 'member-id-' + uuid.uuid4().hex,
        'image_id': 'image-id-' + uuid.uuid4().hex,
        'status': 'pending',
    }

    # Overwrite default attributes if there are some attributes set
    image_member_info.update(attrs)

    return member.Member(**image_member_info)


def create_one_import_info(attrs=None):
    """Create a fake import info.

    :param attrs: A dictionary with all attributes of import info
    :type attrs: dict
    :return: A fake Import object.
    :rtype: `openstack.image.v2.service_info.Import`
    """
    attrs = attrs or {}

    import_info = {
        'import-methods': {
            'description': 'Import methods available.',
            'type': 'array',
            'value': [
                'glance-direct',
                'web-download',
                'glance-download',
                'copy-image',
            ],
        }
    }
    import_info.update(attrs)

    return _service_info.Import(**import_info)


def create_one_stores_info(attrs=None):
    """Create a fake stores info.

    :param attrs: A dictionary with all attributes of stores
    :type attrs: dict
    :return: A fake Store object list.
    :rtype: `openstack.image.v2.service_info.Store`
    """
    attrs = attrs or {}

    stores_info = {
        "stores": [
            {
                "id": "reliable",
                "description": "More expensive store with data redundancy",
            },
            {
                "id": "fast",
                "description": "Provides quick access to your image data",
                "default": True,
            },
            {
                "id": "cheap",
                "description": "Less expensive store for seldom-used images",
            },
        ]
    }
    stores_info.update(attrs)

    return _service_info.Store(**stores_info)


def create_one_task(attrs=None):
    """Create a fake task.

    :param attrs: A dictionary with all attributes of task
    :type attrs: dict
    :return: A fake Task object.
    :rtype: `openstack.image.v2.task.Task`
    """
    attrs = attrs or {}

    # Set default attribute
    task_info = {
        'created_at': '2016-06-29T16:13:07Z',
        'expires_at': '2016-07-01T16:13:07Z',
        'id': str(uuid.uuid4()),
        'input': {
            'image_properties': {
                'container_format': 'ovf',
                'disk_format': 'vhd',
            },
            'import_from': 'https://apps.openstack.org/excellent-image',
            'import_from_format': 'qcow2',
        },
        'message': '',
        'owner': str(uuid.uuid4()),
        'result': {
            'image_id': str(uuid.uuid4()),
        },
        'schema': '/v2/schemas/task',
        'status': random.choice(
            [
                'pending',
                'processing',
                'success',
                'failure',
            ]
        ),
        # though not documented, the API only allows 'import'
        # https://github.com/openstack/glance/blob/24.0.0/glance/api/v2/tasks.py#L186-L190
        'type': 'import',
        'updated_at': '2016-06-29T16:13:07Z',
    }

    # Overwrite default attributes if there are some attributes set
    task_info.update(attrs)

    return task.Task(**task_info)


def create_tasks(attrs=None, count=2):
    """Create multiple fake tasks.

    :param attrs: A dictionary with all attributes of Task
    :type attrs: dict
    :param count: The number of tasks to be faked
    :type count: int
    :return: A list of fake Task objects
    :rtype: list
    """
    tasks = []
    for n in range(0, count):
        tasks.append(create_one_task(attrs))

    return tasks


def create_cache(attrs=None):
    attrs = attrs or {}
    cache_info = {
        'cached_images': [
            {
                'hits': 0,
                'image_id': '1a56983c-f71f-490b-a7ac-6b321a18935a',
                'last_accessed': 1671699579.444378,
                'last_modified': 1671699579.444378,
                'size': 0,
            },
        ],
        'queued_images': [
            '3a4560a1-e585-443e-9b39-553b46ec92d1',
            '6f99bf80-2ee6-47cf-acfe-1f1fabb7e810',
        ],
    }
    cache_info.update(attrs)

    return cache.Cache(**cache_info)


def create_one_metadef_namespace(attrs=None):
    """Create a fake MetadefNamespace member.

    :param attrs: A dictionary with all attributes of metadef_namespace member
    :type attrs: dict
    :return: a list of MetadefNamespace objects
    :rtype: list of `metadef_namespace.MetadefNamespace`
    """
    attrs = attrs or {}

    metadef_namespace_list = {
        'created_at': '2022-08-17T11:30:22Z',
        'display_name': 'Flavor Quota',
        'namespace': 'OS::Compute::Quota',
        'owner': 'admin',
        # 'resource_type_associations': ['OS::Nova::Flavor'],
        # The part that receives the list type factor is not implemented.
        'visibility': 'public',
    }

    # Overwrite default attributes if there are some attributes set
    metadef_namespace_list.update(attrs)
    return metadef_namespace.MetadefNamespace(**metadef_namespace_list)


def create_one_metadef_property(attrs=None):
    attrs = attrs or {}

    metadef_property_list = {
        'name': 'cpu_cores',
        'title': 'vCPU Cores',
        'type': 'integer',
    }

    # Overwrite default attributes if there are some attributes set
    metadef_property_list.update(attrs)
    return metadef_property.MetadefProperty(**metadef_property_list)


def create_one_resource_type(attrs=None):
    """Create a fake MetadefResourceType member.

    :param attrs: A dictionary with all attributes of
        metadef_resource_type member
    :type attrs: dict
    :return: a fake MetadefResourceType object
    :rtype: A `metadef_resource_type.MetadefResourceType`
    """
    attrs = attrs or {}

    metadef_resource_type_info = {
        'name': 'OS::Compute::Quota',
        'properties_target': 'image',
    }

    metadef_resource_type_info.update(attrs)
    return metadef_resource_type.MetadefResourceType(
        **metadef_resource_type_info
    )


def create_resource_types(attrs=None, count=2):
    """Create multiple fake resource types.

    :param attrs: A dictionary with all attributes of
        metadef_resource_type member
    :type attrs: dict
    :return: A list of fake MetadefResourceType objects
    :rtype: list
    """
    metadef_resource_types = []
    for n in range(0, count):
        metadef_resource_types.append(create_one_resource_type(attrs))

    return metadef_resource_types


def create_one_metadef_object(attrs=None):
    """Create a fake MetadefNamespace member.

    :param attrs: A dictionary with all attributes of metadef_namespace member
    :type attrs: dict
    :return: a list of MetadefNamespace objects
    :rtype: list of `metadef_namespace.MetadefNamespace`
    """
    attrs = attrs or {}

    metadef_objects_list = {
        'created_at': '2014-09-19T18:20:56Z',
        'description': 'The CPU limits with control parameters.',
        'name': 'CPU Limits',
        'properties': {
            'quota:cpu_period': {
                'description': 'The enforcement interval',
                'maximum': 1000000,
                'minimum': 1000,
                'title': 'Quota: CPU Period',
                'type': 'integer',
            },
            'quota:cpu_quota': {
                'description': 'The maximum allowed bandwidth',
                'title': 'Quota: CPU Quota',
                'type': 'integer',
            },
            'quota:cpu_shares': {
                'description': 'The proportional weighted',
                'title': 'Quota: CPU Shares',
                'type': 'integer',
            },
        },
        'required': [],
        'schema': '/v2/schemas/metadefs/object',
        'updated_at': '2014-09-19T18:20:56Z',
    }

    # Overwrite default attributes if there are some attributes set
    metadef_objects_list.update(attrs)
    return metadef_object.MetadefObject(**metadef_objects_list)


def create_one_resource_type_association(attrs=None):
    """Create a fake MetadefResourceTypeAssociation.

    :param attrs: A dictionary with all attributes of
        metadef_resource_type_association member
    :type attrs: dict
    :return: A fake MetadefResourceTypeAssociation object
    :rtype: A `metadef_resource_type_association.
        MetadefResourceTypeAssociation`
    """
    attrs = attrs or {}

    metadef_resource_type_association_info = {
        'namespace_name': 'OS::Compute::Quota',
        'name': 'OS::Nova::Flavor',
    }

    metadef_resource_type_association_info.update(attrs)
    return metadef_resource_type.MetadefResourceTypeAssociation(
        **metadef_resource_type_association_info
    )


def create_resource_type_associations(attrs=None, count=2):
    """Create mutiple fake resource type associations/

    :param attrs: A dictionary with all attributes of
        metadef_resource_type_association member
    :type attrs: dict
    :return: A list of fake MetadefResourceTypeAssociation objects
    :rtype: list
    """
    resource_type_associations = []
    for n in range(0, count):
        resource_type_associations.append(
            create_one_resource_type_association(attrs)
        )

    return resource_type_associations



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Compute v2 Console auth token implementations."""

import argparse
from collections.abc import Iterable, Sequence
from typing import Any

from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


def _get_console_connection_columns(
    item: Any,
) -> tuple[tuple[str, ...], tuple[str, ...]]:
    column_map: dict[str, str] = {}
    hidden_columns = ['id', 'location', 'name']
    return utils.get_osc_show_columns_for_sdk_resource(
        item, column_map, hidden_columns
    )


class ShowConsoleConnectionInformation(command.ShowOne):
    _description = _("Show server's remote console connection information")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'token',
            metavar='<token>',
            help=_("Nova console token to lookup"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        compute_client = self.app.client_manager.compute
        auth_token = compute_client.validate_console_auth_token(
            parsed_args.token
        )
        display_columns, columns = _get_console_connection_columns(auth_token)
        data = utils.get_dict_properties(auth_token, columns)

        return (display_columns, data)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.identity.v3 import project as _project
from openstack.identity.v3 import user as _user
from openstack.test import fakes as sdk_fakes
from osc_lib import exceptions
from osc_lib import utils as common_utils

from openstackclient.compute.v2 import server_migration
from openstackclient.tests.unit.compute.v2 import fakes as compute_fakes


class TestListMigration(compute_fakes.TestCompute):
    """Test fetch all migrations."""

    MIGRATION_COLUMNS = [
        'Source Node',
        'Dest Node',
        'Source Compute',
        'Dest Compute',
        'Dest Host',
        'Status',
        'Server UUID',
        'Old Flavor',
        'New Flavor',
        'Created At',
        'Updated At',
    ]

    MIGRATION_FIELDS = [
        'source_node',
        'dest_node',
        'source_compute',
        'dest_compute',
        'dest_host',
        'status',
        'server_id',
        'old_flavor_id',
        'new_flavor_id',
        'created_at',
        'updated_at',
    ]

    def setUp(self):
        super().setUp()

        self.server = compute_fakes.create_one_server()
        self.compute_client.find_server.return_value = self.server

        self.migrations = compute_fakes.create_migrations(count=3)
        self.compute_client.migrations.return_value = self.migrations

        self.data = (
            common_utils.get_item_properties(s, self.MIGRATION_FIELDS)
            for s in self.migrations
        )

        # Get the command object to test
        self.cmd = server_migration.ListMigration(self.app, None)

    def test_server_migration_list_no_options(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {}

        self.compute_client.migrations.assert_called_with(**kwargs)

        self.assertEqual(self.MIGRATION_COLUMNS, columns)
        self.assertEqual(tuple(self.data), tuple(data))

    def test_server_migration_list(self):
        arglist = [
            '--server',
            'server1',
            '--host',
            'host1',
            '--status',
            'migrating',
            '--type',
            'cold-migration',
        ]
        verifylist = [
            ('server', 'server1'),
            ('host', 'host1'),
            ('status', 'migrating'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'status': 'migrating',
            'host': 'host1',
            'instance_uuid': self.server.id,
            'migration_type': 'migration',
        }

        self.compute_client.find_server.assert_called_with(
            'server1', ignore_missing=False
        )
        self.compute_client.migrations.assert_called_with(**kwargs)

        self.assertEqual(self.MIGRATION_COLUMNS, columns)
        self.assertEqual(tuple(self.data), tuple(data))


class TestListMigrationV223(TestListMigration):
    """Test fetch all migrations."""

    MIGRATION_COLUMNS = [
        'Id',
        'Source Node',
        'Dest Node',
        'Source Compute',
        'Dest Compute',
        'Dest Host',
        'Status',
        'Server UUID',
        'Old Flavor',
        'New Flavor',
        'Type',
        'Created At',
        'Updated At',
    ]

    # These are the Migration object fields.
    MIGRATION_FIELDS = [
        'id',
        'source_node',
        'dest_node',
        'source_compute',
        'dest_compute',
        'dest_host',
        'status',
        'server_id',
        'old_flavor_id',
        'new_flavor_id',
        'migration_type',
        'created_at',
        'updated_at',
    ]

    def setUp(self):
        super().setUp()

        self.set_compute_api_version('2.23')

    def test_server_migration_list(self):
        arglist = ['--status', 'migrating']
        verifylist = [('status', 'migrating')]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'status': 'migrating',
        }

        self.compute_client.migrations.assert_called_with(**kwargs)

        self.assertEqual(self.MIGRATION_COLUMNS, columns)
        self.assertEqual(tuple(self.data), tuple(data))


class TestListMigrationV259(TestListMigration):
    """Test fetch all migrations."""

    MIGRATION_COLUMNS = [
        'Id',
        'UUID',
        'Source Node',
        'Dest Node',
        'Source Compute',
        'Dest Compute',
        'Dest Host',
        'Status',
        'Server UUID',
        'Old Flavor',
        'New Flavor',
        'Type',
        'Created At',
        'Updated At',
    ]

    # These are the Migration object fields.
    MIGRATION_FIELDS = [
        'id',
        'uuid',
        'source_node',
        'dest_node',
        'source_compute',
        'dest_compute',
        'dest_host',
        'status',
        'server_id',
        'old_flavor_id',
        'new_flavor_id',
        'migration_type',
        'created_at',
        'updated_at',
    ]

    def setUp(self):
        super().setUp()

        self.set_compute_api_version('2.59')

    def test_server_migration_list(self):
        arglist = [
            '--status',
            'migrating',
            '--limit',
            '1',
            '--marker',
            'test_kp',
            '--changes-since',
            '2019-08-09T08:03:25Z',
        ]
        verifylist = [
            ('status', 'migrating'),
            ('limit', 1),
            ('marker', 'test_kp'),
            ('changes_since', '2019-08-09T08:03:25Z'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'status': 'migrating',
            'limit': 1,
            'paginated': False,
            'marker': 'test_kp',
            'changes_since': '2019-08-09T08:03:25Z',
        }

        self.compute_client.migrations.assert_called_with(**kwargs)

        self.assertEqual(self.MIGRATION_COLUMNS, columns)
        self.assertEqual(tuple(self.data), tuple(data))

    def test_server_migration_list_with_limit_pre_v259(self):
        self.set_compute_api_version('2.58')

        arglist = ['--status', 'migrating', '--limit', '1']
        verifylist = [('status', 'migrating'), ('limit', 1)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        ex = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-compute-api-version 2.59 or greater is required', str(ex)
        )

    def test_server_migration_list_with_marker_pre_v259(self):
        self.set_compute_api_version('2.58')

        arglist = ['--status', 'migrating', '--marker', 'test_kp']
        verifylist = [('status', 'migrating'), ('marker', 'test_kp')]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        ex = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-compute-api-version 2.59 or greater is required', str(ex)
        )

    def test_server_migration_list_with_changes_since_pre_v259(self):
        self.set_compute_api_version('2.58')

        arglist = [
            '--status',
            'migrating',
            '--changes-since',
            '2019-08-09T08:03:25Z',
        ]
        verifylist = [
            ('status', 'migrating'),
            ('changes_since', '2019-08-09T08:03:25Z'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        ex = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-compute-api-version 2.59 or greater is required', str(ex)
        )


class TestListMigrationV266(TestListMigration):
    """Test fetch all migrations by changes-before."""

    MIGRATION_COLUMNS = [
        'Id',
        'UUID',
        'Source Node',
        'Dest Node',
        'Source Compute',
        'Dest Compute',
        'Dest Host',
        'Status',
        'Server UUID',
        'Old Flavor',
        'New Flavor',
        'Type',
        'Created At',
        'Updated At',
    ]

    # These are the Migration object fields.
    MIGRATION_FIELDS = [
        'id',
        'uuid',
        'source_node',
        'dest_node',
        'source_compute',
        'dest_compute',
        'dest_host',
        'status',
        'server_id',
        'old_flavor_id',
        'new_flavor_id',
        'migration_type',
        'created_at',
        'updated_at',
    ]

    def setUp(self):
        super().setUp()

        self.set_compute_api_version('2.66')

    def test_server_migration_list_with_changes_before(self):
        arglist = [
            '--status',
            'migrating',
            '--limit',
            '1',
            '--marker',
            'test_kp',
            '--changes-since',
            '2019-08-07T08:03:25Z',
            '--changes-before',
            '2019-08-09T08:03:25Z',
        ]
        verifylist = [
            ('status', 'migrating'),
            ('limit', 1),
            ('marker', 'test_kp'),
            ('changes_since', '2019-08-07T08:03:25Z'),
            ('changes_before', '2019-08-09T08:03:25Z'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'status': 'migrating',
            'limit': 1,
            'paginated': False,
            'marker': 'test_kp',
            'changes_since': '2019-08-07T08:03:25Z',
            'changes_before': '2019-08-09T08:03:25Z',
        }

        self.compute_client.migrations.assert_called_with(**kwargs)

        self.assertEqual(self.MIGRATION_COLUMNS, columns)
        self.assertEqual(tuple(self.data), tuple(data))

    def test_server_migration_list_with_changes_before_pre_v266(self):
        self.set_compute_api_version('2.65')

        arglist = [
            '--status',
            'migrating',
            '--changes-before',
            '2019-08-09T08:03:25Z',
        ]
        verifylist = [
            ('status', 'migrating'),
            ('changes_before', '2019-08-09T08:03:25Z'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        ex = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-compute-api-version 2.66 or greater is required', str(ex)
        )


class TestListMigrationV280(TestListMigration):
    """Test fetch all migrations by user-id and/or project-id."""

    MIGRATION_COLUMNS = [
        'Id',
        'UUID',
        'Source Node',
        'Dest Node',
        'Source Compute',
        'Dest Compute',
        'Dest Host',
        'Status',
        'Server UUID',
        'Old Flavor',
        'New Flavor',
        'Type',
        'Created At',
        'Updated At',
    ]

    # These are the Migration object fields.
    MIGRATION_FIELDS = [
        'id',
        'uuid',
        'source_node',
        'dest_node',
        'source_compute',
        'dest_compute',
        'dest_host',
        'status',
        'server_id',
        'old_flavor_id',
        'new_flavor_id',
        'migration_type',
        'created_at',
        'updated_at',
    ]

    def setUp(self):
        super().setUp()

        self.project = sdk_fakes.generate_fake_resource(_project.Project)
        self.user = sdk_fakes.generate_fake_resource(_user.User)

        self.identity_sdk_client.find_project.return_value = self.project
        self.identity_sdk_client.find_user.return_value = self.user

        self.set_compute_api_version('2.80')

    def test_server_migration_list_with_project(self):
        arglist = [
            '--status',
            'migrating',
            '--limit',
            '1',
            '--marker',
            'test_kp',
            '--changes-since',
            '2019-08-07T08:03:25Z',
            '--changes-before',
            '2019-08-09T08:03:25Z',
            '--project',
            self.project.id,
        ]
        verifylist = [
            ('status', 'migrating'),
            ('limit', 1),
            ('marker', 'test_kp'),
            ('changes_since', '2019-08-07T08:03:25Z'),
            ('changes_before', '2019-08-09T08:03:25Z'),
            ('project', self.project.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'status': 'migrating',
            'limit': 1,
            'paginated': False,
            'marker': 'test_kp',
            'project_id': self.project.id,
            'changes_since': '2019-08-07T08:03:25Z',
            'changes_before': "2019-08-09T08:03:25Z",
        }

        self.compute_client.migrations.assert_called_with(**kwargs)

        self.MIGRATION_COLUMNS.insert(
            len(self.MIGRATION_COLUMNS) - 2, "Project"
        )
        self.MIGRATION_FIELDS.insert(
            len(self.MIGRATION_FIELDS) - 2, "project_id"
        )
        self.assertEqual(self.MIGRATION_COLUMNS, columns)
        self.assertEqual(tuple(self.data), tuple(data))
        # Clean up global variables MIGRATION_COLUMNS
        self.MIGRATION_COLUMNS.remove('Project')
        # Clean up global variables MIGRATION_FIELDS
        self.MIGRATION_FIELDS.remove('project_id')

    def test_get_migrations_with_project_pre_v280(self):
        self.set_compute_api_version('2.79')

        arglist = [
            '--status',
            'migrating',
            '--changes-before',
            '2019-08-09T08:03:25Z',
            '--project',
            '0c2accde-644a-45fa-8c10-e76debc7fbc3',
        ]
        verifylist = [
            ('status', 'migrating'),
            ('changes_before', '2019-08-09T08:03:25Z'),
            ('project', '0c2accde-644a-45fa-8c10-e76debc7fbc3'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        ex = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-compute-api-version 2.80 or greater is required', str(ex)
        )

    def test_server_migration_list_with_user(self):
        arglist = [
            '--status',
            'migrating',
            '--limit',
            '1',
            '--marker',
            'test_kp',
            '--changes-since',
            '2019-08-07T08:03:25Z',
            '--changes-before',
            '2019-08-09T08:03:25Z',
            '--user',
            self.user.id,
        ]
        verifylist = [
            ('status', 'migrating'),
            ('limit', 1),
            ('marker', 'test_kp'),
            ('changes_since', '2019-08-07T08:03:25Z'),
            ('changes_before', '2019-08-09T08:03:25Z'),
            ('user', self.user.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'status': 'migrating',
            'limit': 1,
            'paginated': False,
            'marker': 'test_kp',
            'user_id': self.user.id,
            'changes_since': '2019-08-07T08:03:25Z',
            'changes_before': "2019-08-09T08:03:25Z",
        }

        self.compute_client.migrations.assert_called_with(**kwargs)

        self.MIGRATION_COLUMNS.insert(len(self.MIGRATION_COLUMNS) - 2, "User")
        self.MIGRATION_FIELDS.insert(len(self.MIGRATION_FIELDS) - 2, "user_id")
        self.assertEqual(self.MIGRATION_COLUMNS, columns)
        self.assertEqual(tuple(self.data), tuple(data))
        # Clean up global variables MIGRATION_COLUMNS
        self.MIGRATION_COLUMNS.remove('User')
        # Clean up global variables MIGRATION_FIELDS
        self.MIGRATION_FIELDS.remove('user_id')

    def test_get_migrations_with_user_pre_v280(self):
        self.set_compute_api_version('2.79')

        arglist = [
            '--status',
            'migrating',
            '--changes-before',
            '2019-08-09T08:03:25Z',
            '--user',
            self.user.id,
        ]
        verifylist = [
            ('status', 'migrating'),
            ('changes_before', '2019-08-09T08:03:25Z'),
            ('user', self.user.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        ex = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-compute-api-version 2.80 or greater is required', str(ex)
        )

    def test_server_migration_list_with_project_and_user(self):
        arglist = [
            '--status',
            'migrating',
            '--limit',
            '1',
            '--changes-since',
            '2019-08-07T08:03:25Z',
            '--changes-before',
            '2019-08-09T08:03:25Z',
            '--project',
            self.project.id,
            '--user',
            self.user.id,
        ]
        verifylist = [
            ('status', 'migrating'),
            ('limit', 1),
            ('changes_since', '2019-08-07T08:03:25Z'),
            ('changes_before', '2019-08-09T08:03:25Z'),
            ('project', self.project.id),
            ('user', self.user.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'status': 'migrating',
            'limit': 1,
            'paginated': False,
            'project_id': self.project.id,
            'user_id': self.user.id,
            'changes_since': '2019-08-07T08:03:25Z',
            'changes_before': "2019-08-09T08:03:25Z",
        }

        self.compute_client.migrations.assert_called_with(**kwargs)

        self.MIGRATION_COLUMNS.insert(
            len(self.MIGRATION_COLUMNS) - 2, "Project"
        )
        self.MIGRATION_FIELDS.insert(
            len(self.MIGRATION_FIELDS) - 2, "project_id"
        )
        self.MIGRATION_COLUMNS.insert(len(self.MIGRATION_COLUMNS) - 2, "User")
        self.MIGRATION_FIELDS.insert(len(self.MIGRATION_FIELDS) - 2, "user_id")
        self.assertEqual(self.MIGRATION_COLUMNS, columns)
        self.assertEqual(tuple(self.data), tuple(data))
        # Clean up global variables MIGRATION_COLUMNS
        self.MIGRATION_COLUMNS.remove('Project')
        self.MIGRATION_FIELDS.remove('project_id')
        self.MIGRATION_COLUMNS.remove('User')
        self.MIGRATION_FIELDS.remove('user_id')

    def test_get_migrations_with_project_and_user_pre_v280(self):
        self.set_compute_api_version('2.79')

        arglist = [
            '--status',
            'migrating',
            '--changes-before',
            '2019-08-09T08:03:25Z',
            '--project',
            self.project.id,
            '--user',
            self.user.id,
        ]
        verifylist = [
            ('status', 'migrating'),
            ('changes_before', '2019-08-09T08:03:25Z'),
            ('project', self.project.id),
            ('user', self.user.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        ex = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-compute-api-version 2.80 or greater is required', str(ex)
        )


class TestServerMigrationShow(compute_fakes.TestCompute):
    def setUp(self):
        super().setUp()

        self.server = compute_fakes.create_one_server()
        self.compute_client.find_server.return_value = self.server

        self.server_migration = compute_fakes.create_one_server_migration()
        self.compute_client.get_server_migration.return_value = (
            self.server_migration
        )
        self.compute_client.server_migrations.return_value = iter(
            [self.server_migration]
        )

        self.columns = (
            'ID',
            'Server UUID',
            'Status',
            'Source Compute',
            'Source Node',
            'Dest Compute',
            'Dest Host',
            'Dest Node',
            'Memory Total Bytes',
            'Memory Processed Bytes',
            'Memory Remaining Bytes',
            'Disk Total Bytes',
            'Disk Processed Bytes',
            'Disk Remaining Bytes',
            'Created At',
            'Updated At',
        )

        self.data = (
            self.server_migration.id,
            self.server_migration.server_id,
            self.server_migration.status,
            self.server_migration.source_compute,
            self.server_migration.source_node,
            self.server_migration.dest_compute,
            self.server_migration.dest_host,
            self.server_migration.dest_node,
            self.server_migration.memory_total_bytes,
            self.server_migration.memory_processed_bytes,
            self.server_migration.memory_remaining_bytes,
            self.server_migration.disk_total_bytes,
            self.server_migration.disk_processed_bytes,
            self.server_migration.disk_remaining_bytes,
            self.server_migration.created_at,
            self.server_migration.updated_at,
        )

        # Get the command object to test
        self.cmd = server_migration.ShowMigration(self.app, None)

    def _test_server_migration_show(self):
        arglist = [
            self.server.id,
            '2',  # arbitrary migration ID
        ]
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

        self.compute_client.find_server.assert_called_with(
            self.server.id, ignore_missing=False
        )
        self.compute_client.get_server_migration.assert_called_with(
            self.server.id, '2', ignore_missing=False
        )

    def test_server_migration_show(self):
        self.set_compute_api_version('2.24')

        self._test_server_migration_show()

    def test_server_migration_show_v259(self):
        self.set_compute_api_version('2.59')

        self.columns += ('UUID',)
        self.data += (self.server_migration.uuid,)

        self._test_server_migration_show()

    def test_server_migration_show_v280(self):
        self.set_compute_api_version('2.80')

        self.columns += ('UUID', 'User ID', 'Project ID')
        self.data += (
            self.server_migration.uuid,
            self.server_migration.user_id,
            self.server_migration.project_id,
        )

        self._test_server_migration_show()

    def test_server_migration_show_pre_v224(self):
        self.set_compute_api_version('2.23')

        arglist = [
            self.server.id,
            '2',  # arbitrary migration ID
        ]
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        ex = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-compute-api-version 2.24 or greater is required', str(ex)
        )

    def test_server_migration_show_by_uuid(self):
        self.set_compute_api_version('2.59')

        self.compute_client.server_migrations.return_value = iter(
            [self.server_migration]
        )

        self.columns += ('UUID',)
        self.data += (self.server_migration.uuid,)

        arglist = [
            self.server.id,
            self.server_migration.uuid,  # arbitrary migration UUID
        ]
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

        self.compute_client.find_server.assert_called_with(
            self.server.id, ignore_missing=False
        )
        self.compute_client.server_migrations.assert_called_with(
            self.server.id
        )
        self.compute_client.get_server_migration.assert_not_called()

    def test_server_migration_show_by_uuid_no_matches(self):
        self.set_compute_api_version('2.59')

        self.compute_client.server_migrations.return_value = iter([])

        arglist = [
            self.server.id,
            '69f95745-bfe3-4302-90f7-5b0022cba1ce',  # arbitrary migration UUID
        ]
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        ex = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            'In-progress live migration 69f95745-bfe3-4302-90f7-5b0022cba1ce',
            str(ex),
        )

    def test_server_migration_show_by_uuid_pre_v259(self):
        self.set_compute_api_version('2.58')

        arglist = [
            self.server.id,
            '69f95745-bfe3-4302-90f7-5b0022cba1ce',  # arbitrary migration UUID
        ]
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        ex = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-compute-api-version 2.59 or greater is required', str(ex)
        )

    def test_server_migration_show_invalid_id(self):
        self.set_compute_api_version('2.24')

        arglist = [
            self.server.id,
            'foo',  # invalid migration ID
        ]
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        ex = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            'The <migration> argument must be an ID or UUID', str(ex)
        )


class TestServerMigrationAbort(compute_fakes.TestCompute):
    def setUp(self):
        super().setUp()

        self.server = compute_fakes.create_one_server()

        # Return value for utils.find_resource for server.
        self.compute_client.find_server.return_value = self.server

        # Get the command object to test
        self.cmd = server_migration.AbortMigration(self.app, None)

    def test_migration_abort(self):
        self.set_compute_api_version('2.24')

        arglist = [
            self.server.id,
            '2',  # arbitrary migration ID
        ]
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.compute_client.find_server.assert_called_with(
            self.server.id, ignore_missing=False
        )
        self.compute_client.abort_server_migration.assert_called_with(
            '2', self.server.id, ignore_missing=False
        )
        self.assertIsNone(result)

    def test_migration_abort_pre_v224(self):
        self.set_compute_api_version('2.23')

        arglist = [
            self.server.id,
            '2',  # arbitrary migration ID
        ]
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        ex = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-compute-api-version 2.24 or greater is required', str(ex)
        )

    def test_server_migration_abort_by_uuid(self):
        self.set_compute_api_version('2.59')

        self.server_migration = compute_fakes.create_one_server_migration()
        self.compute_client.server_migrations.return_value = iter(
            [self.server_migration]
        )

        arglist = [
            self.server.id,
            self.server_migration.uuid,  # arbitrary migration UUID
        ]
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.compute_client.find_server.assert_called_with(
            self.server.id, ignore_missing=False
        )
        self.compute_client.server_migrations.assert_called_with(
            self.server.id
        )
        self.compute_client.abort_server_migration.assert_called_with(
            self.server_migration.id, self.server.id, ignore_missing=False
        )
        self.assertIsNone(result)

    def test_server_migration_abort_by_uuid_no_matches(self):
        self.set_compute_api_version('2.59')

        self.compute_client.server_migrations.return_value = iter([])

        arglist = [
            self.server.id,
            '69f95745-bfe3-4302-90f7-5b0022cba1ce',  # arbitrary migration UUID
        ]
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        ex = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            'In-progress live migration 69f95745-bfe3-4302-90f7-5b0022cba1ce',
            str(ex),
        )

    def test_server_migration_abort_by_uuid_pre_v259(self):
        self.set_compute_api_version('2.58')

        arglist = [
            self.server.id,
            '69f95745-bfe3-4302-90f7-5b0022cba1ce',  # arbitrary migration UUID
        ]
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        ex = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-compute-api-version 2.59 or greater is required', str(ex)
        )


class TestServerMigrationForceComplete(compute_fakes.TestCompute):
    def setUp(self):
        super().setUp()

        self.server = compute_fakes.create_one_server()

        # Return value for utils.find_resource for server.
        self.compute_client.find_server.return_value = self.server

        # Get the command object to test
        self.cmd = server_migration.ForceCompleteMigration(self.app, None)

    def test_migration_force_complete(self):
        self.set_compute_api_version('2.22')

        arglist = [
            self.server.id,
            '2',  # arbitrary migration ID
        ]
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.compute_client.find_server.assert_called_with(
            self.server.id, ignore_missing=False
        )
        self.compute_client.force_complete_server_migration.assert_called_with(
            '2', self.server.id
        )
        self.assertIsNone(result)

    def test_migration_force_complete_pre_v222(self):
        self.set_compute_api_version('2.21')

        arglist = [
            self.server.id,
            '2',  # arbitrary migration ID
        ]
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        ex = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-compute-api-version 2.22 or greater is required', str(ex)
        )

    def test_server_migration_force_complete_by_uuid(self):
        self.set_compute_api_version('2.59')

        self.server_migration = compute_fakes.create_one_server_migration()
        self.compute_client.server_migrations.return_value = iter(
            [self.server_migration]
        )

        arglist = [
            self.server.id,
            self.server_migration.uuid,  # arbitrary migration UUID
        ]
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.compute_client.find_server.assert_called_with(
            self.server.id, ignore_missing=False
        )
        self.compute_client.server_migrations.assert_called_with(
            self.server.id
        )
        self.compute_client.force_complete_server_migration.assert_called_with(
            self.server_migration.id, self.server.id
        )
        self.assertIsNone(result)

    def test_server_migration_force_complete_by_uuid_no_matches(self):
        self.set_compute_api_version('2.59')

        self.compute_client.server_migrations.return_value = iter([])

        arglist = [
            self.server.id,
            '69f95745-bfe3-4302-90f7-5b0022cba1ce',  # arbitrary migration UUID
        ]
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        ex = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            'In-progress live migration 69f95745-bfe3-4302-90f7-5b0022cba1ce',
            str(ex),
        )

    def test_server_migration_force_complete_by_uuid_pre_v259(self):
        self.set_compute_api_version('2.58')

        arglist = [
            self.server.id,
            '69f95745-bfe3-4302-90f7-5b0022cba1ce',  # arbitrary migration UUID
        ]
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        ex = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-compute-api-version 2.59 or greater is required', str(ex)
        )



#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from openstack import exceptions as sdk_exceptions
from openstack.identity.v3 import service as _service
from openstack.test import fakes as sdk_fakes
from osc_lib import exceptions

from openstackclient.identity.v3 import service
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes


class TestServiceCreate(identity_fakes.TestIdentity):
    columns = (
        'id',
        'name',
        'type',
        'enabled',
        'description',
    )

    def setUp(self):
        super().setUp()

        self.service = sdk_fakes.generate_fake_resource(_service.Service)

        self.datalist = (
            self.service.id,
            self.service.name,
            self.service.type,
            True,
            self.service.description,
        )
        self.identity_sdk_client.create_service.return_value = self.service

        # Get the command object to test
        self.cmd = service.CreateService(self.app, None)

    def test_service_create_name(self):
        arglist = [
            '--name',
            self.service.name,
            self.service.type,
        ]
        verifylist = [
            ('name', self.service.name),
            ('description', None),
            ('is_enabled', True),
            ('type', self.service.type),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.create_service.assert_called_with(
            name=self.service.name,
            type=self.service.type,
            description=None,
            is_enabled=True,
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_service_create_description(self):
        arglist = [
            '--description',
            self.service.description,
            self.service.type,
        ]
        verifylist = [
            ('name', None),
            ('description', self.service.description),
            ('is_enabled', True),
            ('type', self.service.type),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.create_service.assert_called_with(
            name=None,
            type=self.service.type,
            description=self.service.description,
            is_enabled=True,
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_service_create_enable(self):
        arglist = [
            '--enable',
            self.service.type,
        ]
        verifylist = [
            ('name', None),
            ('description', None),
            ('is_enabled', True),
            ('type', self.service.type),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.create_service.assert_called_with(
            name=None,
            type=self.service.type,
            description=None,
            is_enabled=True,
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_service_create_disable(self):
        arglist = [
            '--disable',
            self.service.type,
        ]
        verifylist = [
            ('name', None),
            ('description', None),
            ('is_enabled', False),
            ('type', self.service.type),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.create_service.assert_called_with(
            name=None,
            type=self.service.type,
            description=None,
            is_enabled=False,
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)


class TestServiceDelete(identity_fakes.TestIdentity):
    service = sdk_fakes.generate_fake_resource(_service.Service)

    def setUp(self):
        super().setUp()

        self.identity_sdk_client.get_service.side_effect = (
            sdk_exceptions.ResourceNotFound
        )
        self.identity_sdk_client.find_service.return_value = self.service
        self.identity_sdk_client.delete_service.return_value = None

        # Get the command object to test
        self.cmd = service.DeleteService(self.app, None)

    def test_service_delete_no_options(self):
        arglist = [
            self.service.name,
        ]
        verifylist = [
            ('service', [self.service.name]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.delete_service.assert_called_with(
            self.service.id,
        )
        self.assertIsNone(result)


class TestServiceList(identity_fakes.TestIdentity):
    service = sdk_fakes.generate_fake_resource(_service.Service)

    def setUp(self):
        super().setUp()

        self.identity_sdk_client.services.return_value = [self.service]

        # Get the command object to test
        self.cmd = service.ListService(self.app, None)

    def test_service_list_no_options(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.services.assert_called_with()

        collist = ('ID', 'Name', 'Type')
        self.assertEqual(collist, columns)
        datalist = (
            (
                self.service.id,
                self.service.name,
                self.service.type,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_service_list_long(self):
        arglist = [
            '--long',
        ]
        verifylist = [
            ('long', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.services.assert_called_with()

        collist = ('ID', 'Name', 'Type', 'Description', 'Enabled')
        self.assertEqual(collist, columns)
        datalist = (
            (
                self.service.id,
                self.service.name,
                self.service.type,
                self.service.description,
                True,
            ),
        )
        self.assertEqual(datalist, tuple(data))


class TestServiceSet(identity_fakes.TestIdentity):
    service = sdk_fakes.generate_fake_resource(_service.Service)

    def setUp(self):
        super().setUp()

        self.identity_sdk_client.get_service.side_effect = (
            sdk_exceptions.ResourceNotFound
        )
        self.identity_sdk_client.find_service.return_value = self.service
        self.identity_sdk_client.update_service.return_value = self.service

        # Get the command object to test
        self.cmd = service.SetService(self.app, None)

    def test_service_set_no_options(self):
        arglist = [
            self.service.name,
        ]
        verifylist = [
            ('type', None),
            ('name', None),
            ('description', None),
            ('is_enabled', None),
            ('service', self.service.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)

    def test_service_set_type(self):
        arglist = [
            '--type',
            self.service.type,
            self.service.name,
        ]
        verifylist = [
            ('type', self.service.type),
            ('name', None),
            ('description', None),
            ('is_enabled', None),
            ('service', self.service.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'type': self.service.type,
        }
        self.identity_sdk_client.update_service.assert_called_with(
            self.service.id, **kwargs
        )
        self.assertIsNone(result)

    def test_service_set_name(self):
        arglist = [
            '--name',
            self.service.name,
            self.service.name,
        ]
        verifylist = [
            ('type', None),
            ('name', self.service.name),
            ('description', None),
            ('is_enabled', None),
            ('service', self.service.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.service.name,
        }
        self.identity_sdk_client.update_service.assert_called_with(
            self.service.id, **kwargs
        )
        self.assertIsNone(result)

    def test_service_set_description(self):
        arglist = [
            '--description',
            self.service.description,
            self.service.name,
        ]
        verifylist = [
            ('type', None),
            ('name', None),
            ('description', self.service.description),
            ('is_enabled', None),
            ('service', self.service.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'description': self.service.description,
        }
        self.identity_sdk_client.update_service.assert_called_with(
            self.service.id, **kwargs
        )
        self.assertIsNone(result)

    def test_service_set_enable(self):
        arglist = [
            '--enable',
            self.service.name,
        ]
        verifylist = [
            ('type', None),
            ('name', None),
            ('description', None),
            ('is_enabled', True),
            ('service', self.service.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'is_enabled': True,
        }
        self.identity_sdk_client.update_service.assert_called_with(
            self.service.id, **kwargs
        )
        self.assertIsNone(result)

    def test_service_set_disable(self):
        arglist = [
            '--disable',
            self.service.name,
        ]
        verifylist = [
            ('type', None),
            ('name', None),
            ('description', None),
            ('is_enabled', False),
            ('service', self.service.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'is_enabled': False,
        }
        self.identity_sdk_client.update_service.assert_called_with(
            self.service.id, **kwargs
        )
        self.assertIsNone(result)


class TestServiceShow(identity_fakes.TestIdentity):
    service = sdk_fakes.generate_fake_resource(_service.Service)

    def setUp(self):
        super().setUp()

        self.identity_sdk_client.get_service.side_effect = (
            sdk_exceptions.ResourceNotFound
        )
        self.identity_sdk_client.find_service.return_value = self.service

        # Get the command object to test
        self.cmd = service.ShowService(self.app, None)

    def test_service_show(self):
        arglist = [
            self.service.name,
        ]
        verifylist = [
            ('service', self.service.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.find_service.assert_called_with(
            self.service.name, ignore_missing=False
        )

        collist = ('id', 'name', 'type', 'enabled', 'description')
        self.assertEqual(collist, columns)
        datalist = (
            self.service.id,
            self.service.name,
            self.service.type,
            True,
            self.service.description,
        )
        self.assertEqual(datalist, data)

    def test_service_show_nounique(self):
        self.identity_sdk_client.find_service.side_effect = (
            sdk_exceptions.DuplicateResource(None)
        )
        arglist = [
            'nounique_service',
        ]
        verifylist = [
            ('service', 'nounique_service'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual(
                "DuplicateResource",
                str(e),
            )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest import mock

from osc_lib.cli import format_columns
from osc_lib import exceptions
from osc_lib import utils as common_utils

from openstackclient.compute.v2 import server_backup
from openstackclient.tests.unit.compute.v2 import fakes as compute_fakes
from openstackclient.tests.unit.image.v2 import fakes as image_fakes


class TestServerBackupCreate(compute_fakes.TestCompute):
    def image_columns(self, image):
        columnlist = (
            'id',
            'name',
            'owner',
            'protected',
            'status',
            'tags',
            'visibility',
        )
        return columnlist

    def image_data(self, image):
        datalist = (
            image['id'],
            image['name'],
            image['owner_id'],
            image['is_protected'],
            'active',
            format_columns.ListColumn(image.get('tags')),
            image['visibility'],
        )
        return datalist

    def setUp(self):
        super().setUp()

        self.server = compute_fakes.create_one_server()
        self.compute_client.find_server.return_value = self.server

        self.image = image_fakes.create_one_image(
            {'name': self.server.name, 'status': 'active'}
        )
        self.image_client.find_image.return_value = self.image

        # Get the command object to test
        self.cmd = server_backup.CreateServerBackup(self.app, None)

    def test_server_backup_defaults(self):
        arglist = [
            self.server.id,
        ]
        verifylist = [
            ('name', None),
            ('type', None),
            ('rotate', None),
            ('wait', False),
            ('server', self.server.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        self.compute_client.backup_server.assert_called_with(
            self.server.id,
            self.server.name,
            '',
            1,
        )

        self.assertEqual(self.image_columns(self.image), columns)
        self.assertCountEqual(self.image_data(self.image), data)

    def test_server_backup_create_options(self):
        arglist = [
            '--name',
            'image',
            '--type',
            'daily',
            '--rotate',
            '2',
            self.server.id,
        ]
        verifylist = [
            ('name', 'image'),
            ('type', 'daily'),
            ('rotate', 2),
            ('server', self.server.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        self.compute_client.backup_server.assert_called_with(
            self.server.id,
            'image',
            'daily',
            2,
        )

        self.assertEqual(self.image_columns(self.image), columns)
        self.assertCountEqual(self.image_data(self.image), data)

    @mock.patch.object(common_utils, 'wait_for_status', return_value=False)
    def test_server_backup_wait_fail(self, mock_wait_for_status):
        self.image_client.get_image.return_value = self.image

        arglist = [
            '--name',
            'image',
            '--type',
            'daily',
            '--wait',
            self.server.id,
        ]
        verifylist = [
            ('name', 'image'),
            ('type', 'daily'),
            ('wait', True),
            ('server', self.server.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args,
        )

        self.compute_client.backup_server.assert_called_with(
            self.server.id,
            'image',
            'daily',
            1,
        )

        mock_wait_for_status.assert_called_once_with(
            self.image_client.get_image, self.image.id, callback=mock.ANY
        )

    @mock.patch.object(common_utils, 'wait_for_status', return_value=True)
    def test_server_backup_wait_ok(self, mock_wait_for_status):
        self.image_client.get_image.side_effect = (self.image,)

        arglist = [
            '--name',
            'image',
            '--type',
            'daily',
            '--wait',
            self.server.id,
        ]
        verifylist = [
            ('name', 'image'),
            ('type', 'daily'),
            ('wait', True),
            ('server', self.server.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        self.compute_client.backup_server.assert_called_with(
            self.server.id,
            'image',
            'daily',
            1,
        )

        mock_wait_for_status.assert_called_once_with(
            self.image_client.get_image, self.image.id, callback=mock.ANY
        )

        self.assertEqual(self.image_columns(self.image), columns)
        self.assertCountEqual(self.image_data(self.image), data)



# Copyright 2023 Red Hat.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest.mock import call

from openstack import exceptions as sdk_exceptions
from osc_lib import exceptions

from openstackclient.image.v2 import cache
from openstackclient.tests.unit.image.v2 import fakes


class TestCacheList(fakes.TestImagev2):
    _cache = fakes.create_cache()
    columns = [
        "ID",
        "State",
        "Last Accessed (UTC)",
        "Last Modified (UTC)",
        "Size",
        "Hits",
    ]

    cache_list = cache._format_image_cache(dict(fakes.create_cache()))
    datalist = (
        (
            image['image_id'],
            image['state'],
            image['last_accessed'],
            image['last_modified'],
            image['size'],
            image['hits'],
        )
        for image in cache_list
    )

    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.image_client.get_image_cache.return_value = self._cache
        self.cmd = cache.ListCachedImage(self.app, None)

    def test_image_cache_list(self):
        arglist = []
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)

        self.image_client.get_image_cache.assert_called()
        self.assertEqual(self.columns, columns)
        self.assertEqual(tuple(self.datalist), tuple(data))


class TestQueueCache(fakes.TestImagev2):
    def setUp(self):
        super().setUp()

        self.image_client.queue_image.return_value = None
        self.cmd = cache.QueueCachedImage(self.app, None)

    def test_cache_queue(self):
        images = fakes.create_images(count=1)
        arglist = [
            images[0].id,
        ]

        verifylist = [
            ('images', [images[0].id]),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.image_client.find_image.side_effect = images

        self.cmd.take_action(parsed_args)

        self.image_client.queue_image.assert_called_once_with(images[0].id)

    def test_cache_queue_multiple_images(self):
        images = fakes.create_images(count=3)
        arglist = [i.id for i in images]

        verifylist = [
            ('images', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.image_client.find_image.side_effect = images

        self.cmd.take_action(parsed_args)
        calls = [call(i.id) for i in images]
        self.image_client.queue_image.assert_has_calls(calls)


class TestCacheDelete(fakes.TestImagev2):
    def setUp(self):
        super().setUp()

        self.image_client.cache_delete_image.return_value = None
        self.cmd = cache.DeleteCachedImage(self.app, None)

    def test_cache_delete(self):
        images = fakes.create_images(count=1)
        arglist = [
            images[0].id,
        ]

        verifylist = [
            ('images', [images[0].id]),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.image_client.find_image.side_effect = images

        self.cmd.take_action(parsed_args)

        self.image_client.find_image.assert_called_once_with(
            images[0].id, ignore_missing=False
        )
        self.image_client.cache_delete_image.assert_called_once_with(
            images[0].id
        )

    def test_cache_delete_multiple_images(self):
        images = fakes.create_images(count=3)
        arglist = [i.id for i in images]

        verifylist = [
            ('images', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.image_client.find_image.side_effect = images

        self.cmd.take_action(parsed_args)
        calls = [call(i.id) for i in images]
        self.image_client.cache_delete_image.assert_has_calls(calls)

    def test_cache_delete_multiple_images_exception(self):
        images = fakes.create_images(count=2)
        arglist = [
            images[0].id,
            images[1].id,
            'x-y-x',
        ]
        verifylist = [
            ('images', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        ret_find = [images[0], images[1], sdk_exceptions.ResourceNotFound()]

        self.image_client.find_image.side_effect = ret_find

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        calls = [call(i.id) for i in images]
        self.image_client.cache_delete_image.assert_has_calls(calls)


class TestCacheClear(fakes.TestImagev2):
    def setUp(self):
        super().setUp()

        self.image_client.clear_cache.return_value = None
        self.cmd = cache.ClearCachedImage(self.app, None)

    def test_cache_clear_no_option(self):
        arglist = []

        verifylist = [('target', 'both')]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        self.assertIsNone(
            self.image_client.clear_cache.assert_called_with('both')
        )

    def test_cache_clear_queue_option(self):
        arglist = ['--queue']

        verifylist = [('target', 'queue')]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        self.image_client.clear_cache.assert_called_once_with('queue')

    def test_cache_clear_cache_option(self):
        arglist = ['--cache']

        verifylist = [('target', 'cache')]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        self.image_client.clear_cache.assert_called_once_with('cache')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import argparse
from collections.abc import Iterable, Sequence
from typing import Any

from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


class AddBgpSpeakerToDRAgent(command.Command):
    """Add a BGP speaker to a dynamic routing agent"""

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'dragent_id',
            metavar='<agent-id>',
            help=_("ID of the dynamic routing agent"),
        )
        parser.add_argument(
            'bgp_speaker',
            metavar='<bgp-speaker>',
            help=_("ID or name of the BGP speaker"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        speaker_id = client.find_bgp_speaker(
            parsed_args.bgp_speaker, ignore_missing=False
        ).id
        client.add_bgp_speaker_to_dragent(parsed_args.dragent_id, speaker_id)


class RemoveBgpSpeakerFromDRAgent(command.Command):
    """Removes a BGP speaker from a dynamic routing agent"""

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'dragent_id',
            metavar='<agent-id>',
            help=_("ID of the dynamic routing agent"),
        )
        parser.add_argument(
            'bgp_speaker',
            metavar='<bgp-speaker>',
            help=_("ID or name of the BGP speaker"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        speaker_id = client.find_bgp_speaker(
            parsed_args.bgp_speaker, ignore_missing=False
        ).id
        client.remove_bgp_speaker_from_dragent(
            parsed_args.dragent_id, speaker_id
        )


class ListDRAgent(command.Lister):
    """List dynamic routing agents"""

    resource = 'agent'
    list_columns = ['id', 'host', 'admin_state_up', 'alive']
    unknown_parts_flag = False

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--bgp-speaker',
            metavar='<bgp-speaker>',
            help=_(
                "List dynamic routing agents hosting a "
                "BGP speaker (name or ID)"
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        data: Iterable[Any]
        if parsed_args.bgp_speaker is not None:
            speaker_id = client.find_bgp_speaker(
                parsed_args.bgp_speaker, ignore_missing=False
            ).id
            data = client.get_bgp_dragents_hosting_speaker(speaker_id)
        else:
            data = client.agents(agent_type='BGP dynamic routing agent')
        columns = (
            'id',
            'agent_type',
            'host',
            'availability_zone',
            'is_alive',
            'is_admin_state_up',
            'binary',
        )
        column_headers = (
            'ID',
            'Agent Type',
            'Host',
            'Availability Zone',
            'Alive',
            'State',
            'Binary',
        )
        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                )
                for s in data
            ),
        )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import argparse
import logging
from collections.abc import Iterable, Sequence
from typing import Any

from manilaclient.common.apiclient import utils as apiutils
from osc_lib import exceptions
from osc_lib import utils as oscutils

from openstackclient import command
from openstackclient.common import envvars
from openstackclient.i18n import _
from openstackclient.share import utils

LOG = logging.getLogger(__name__)

ATTRIBUTES = [
    'id',
    'name',
    'share_types',
    'visibility',
    'is_default',
    'group_specs',
]


class CreateShareGroupType(command.ShowOne):
    """Create new share group type."""

    _description = _("Create new share group type")

    log = logging.getLogger(__name__ + ".CreateShareGroupType")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar="<name>",
            default=None,
            help=_('Share group type name'),
        )
        parser.add_argument(
            "share_types",
            metavar="<share-types>",
            nargs="+",
            default=None,
            help=_(
                "List of share type names or IDs. Example:"
                " my-share-type-1 my-share-type-2"
            ),
        )
        parser.add_argument(
            "--group-specs",
            type=str,
            nargs='*',
            metavar='<key=value>',
            default=None,
            help=_(
                "Share Group type extra specs by key and value."
                " OPTIONAL: Default=None. Example:"
                " --group-specs consistent_snapshot_support=host."
            ),
        )
        parser.add_argument(
            '--public',
            metavar="<public>",
            default=True,
            help=_('Make type accessible to the public (default true).'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], tuple[Any, ...]]:
        share_client = self.app.client_manager.share

        kwargs = {'name': parsed_args.name}

        share_types_list = []
        for share_type in parsed_args.share_types:
            try:
                share_type_obj = apiutils.find_resource(
                    share_client.share_types, share_type
                )

                share_types_list.append(share_type_obj.name)
            except Exception as e:
                msg = _(
                    "Failed to find the share type with name or ID "
                    "'%(share_type)s': %(e)s"
                )
                raise exceptions.CommandError(
                    msg % {'share_type': share_type, 'e': e}
                )

        kwargs['share_types'] = share_types_list

        if parsed_args.public:
            kwargs['is_public'] = envvars.bool_from_str(parsed_args.public)

        group_specs: dict[str, Any] = {}
        if parsed_args.group_specs:
            for item in parsed_args.group_specs:
                group_specs = utils.extract_group_specs(group_specs, [item])

        kwargs['group_specs'] = group_specs

        share_group_type = share_client.share_group_types.create(**kwargs)

        formatter = parsed_args.formatter

        formatted_group_type = utils.format_share_group_type(
            share_group_type, formatter
        )

        return (
            ATTRIBUTES,
            oscutils.get_dict_properties(formatted_group_type, ATTRIBUTES),
        )


class DeleteShareGroupType(command.Command):
    """Delete a share group type."""

    _description = _("Delete a share group type")

    log = logging.getLogger(__name__ + ".DeleteShareGroupType")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'share_group_types',
            metavar="<share-group-types>",
            nargs="+",
            help=_("Name or ID of the share group type(s) to delete"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        share_client = self.app.client_manager.share
        result = 0

        for share_group_type in parsed_args.share_group_types:
            try:
                share_group_type_obj = apiutils.find_resource(
                    share_client.share_group_types, share_group_type
                )

                share_client.share_group_types.delete(share_group_type_obj)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete share group type with "
                        "name or ID '%(share_group_type)s': %(e)s"
                    ),
                    {'share_group_type': share_group_type, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.share_group_types)
            msg = _(
                "%(result)s of %(total)s share group types failed to delete."
            ) % {'result': result, 'total': total}
            raise exceptions.CommandError(msg)


class ListShareGroupType(command.Lister):
    """List Share Group Types."""

    _description = _("List share types")

    log = logging.getLogger(__name__ + ".ListShareGroupType")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--all',
            action='store_true',
            default=False,
            help=_(
                'Display all share group types whether public or private. '
                'Default=False. (Admin only)'
            ),
        )
        parser.add_argument(
            '--group-specs',
            type=str,
            nargs='*',
            metavar='<key=value>',
            default=None,
            help=_('Filter share group types with group specs (key=value).'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        share_client = self.app.client_manager.share

        search_opts = {}
        if parsed_args.group_specs:
            search_opts = {
                'group_specs': utils.extract_group_specs(
                    extra_specs={}, specs_to_add=parsed_args.group_specs
                )
            }

        formatter = parsed_args.formatter

        share_group_types = share_client.share_group_types.list(
            search_opts=search_opts, show_all=parsed_args.all
        )

        formatted_types = []
        for share_group_type in share_group_types:
            formatted_types.append(
                utils.format_share_group_type(share_group_type, formatter)
            )

        column_headers = utils.format_column_headers(ATTRIBUTES)
        values = (
            oscutils.get_dict_properties(sgt, ATTRIBUTES)
            for sgt in formatted_types
        )

        return (column_headers, values)


class ShowShareGroupType(command.ShowOne):
    """Show Share Group Types."""

    _description = _("Show share group types")

    log = logging.getLogger(__name__ + ".ShowShareGroupType")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'share_group_type',
            metavar="<share-group-type>",
            help=_("Name or ID of the share group type to show"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], tuple[Any, ...]]:
        share_client = self.app.client_manager.share

        share_group_type = apiutils.find_resource(
            share_client.share_group_types, parsed_args.share_group_type
        )

        share_group_type_obj = share_client.share_group_types.get(
            share_group_type
        )

        formatter = parsed_args.formatter

        formatted_group_type = utils.format_share_group_type(
            share_group_type_obj, formatter
        )

        return (
            ATTRIBUTES,
            oscutils.get_dict_properties(formatted_group_type, ATTRIBUTES),
        )


class SetShareGroupType(command.Command):
    """Set share type properties."""

    _description = _("Set share group type properties")

    log = logging.getLogger(__name__ + ".SetShareGroupType")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'share_group_type',
            metavar="<share-group-type>",
            help=_("Name or ID of the share group type to modify"),
        )
        parser.add_argument(
            "--group-specs",
            type=str,
            nargs='*',
            metavar='<key=value>',
            default=None,
            help=_(
                "Extra specs key and value of share group type that will be"
                " used for share type creation. OPTIONAL: Default=None."
                " Example: --group-specs consistent-snapshot-support=True"
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        share_client = self.app.client_manager.share

        try:
            share_group_type_obj = apiutils.find_resource(
                share_client.share_group_types, parsed_args.share_group_type
            )
        except Exception as e:
            msg = _(
                "Failed to find the share group type with "
                "name or ID '%(share_group_type)s': %(e)s"
            )
            raise exceptions.CommandError(
                msg
                % {'share_group_type': parsed_args.share_group_type, 'e': e}
            )

        if parsed_args.group_specs:
            group_specs = utils.extract_group_specs(
                extra_specs={}, specs_to_add=parsed_args.group_specs
            )
            try:
                share_group_type_obj.set_keys(group_specs)
            except Exception as e:
                raise exceptions.CommandError(
                    f"Failed to set share group type key: {e}"
                )


class UnsetShareGroupType(command.Command):
    """Unset share group type extra specs."""

    _description = _("Unset share group type extra specs")

    log = logging.getLogger(__name__ + ".UnsetShareGroupType")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'share_group_type',
            metavar="<share-group-type>",
            help=_("Name or ID of the share grouptype to modify"),
        )
        parser.add_argument(
            'group_specs',
            metavar='<key>',
            nargs='+',
            help=_('Remove group specs from this share group type'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        share_client = self.app.client_manager.share

        try:
            share_group_type_obj = apiutils.find_resource(
                share_client.share_group_types, parsed_args.share_group_type
            )
        except Exception as e:
            msg = _(
                "Failed to find the share group type with "
                "name or ID '%(share_group_type)s': %(e)s"
            )
            raise exceptions.CommandError(
                msg
                % {'share_group_type': parsed_args.share_group_type, 'e': e}
            )

        if parsed_args.group_specs:
            try:
                share_group_type_obj.unset_keys(parsed_args.group_specs)
            except Exception as e:
                raise exceptions.CommandError(
                    f"Failed to remove share type group extra spec: {e}"
                )
